/*=========================================================================

						 >>> WonderGUI <<<

  This file is part of Tord Jansson's WonderGUI Graphics Toolkit
  and copyright (c) Tord Jansson, Sweden [tord.jansson@gmail.com].

							-----------

  The WonderGUI Graphics Toolkit is free software; you can redistribute
  this file and/or modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

							-----------

  The WonderGUI Graphics Toolkit is also available for use in commercial
  closed-source projects under a separate license. Interested parties
  should contact Tord Jansson [tord.jansson@gmail.com] for details.

=========================================================================*/

#include <wg_rootpanel.h>
#include <wg_base.h>
#include <wg_container.h>
#include <wg_boxskin.h>
#include <wg_inputhandler.h>
#include <wg_util.h>
#include <wg_internal.h>

#include <new>


#include <wg_msgrouter.h>


namespace wg
{
	using namespace Util;

	const TypeInfo RootPanel::TYPEINFO = { "RootPanel", &Object::TYPEINFO };


	//____ constructor ____________________________________________________________

	RootPanel::RootPanel() : slot(this)
	{
		m_bVisible = true;
		m_bHasGeo = false;

		m_bDebugMode = false;

		BoxSkin_p pDebugOverlay = BoxSkin::create( 1, Color(255,0,0,128), Color(255,0,0,128) );
		pDebugOverlay->setColors( StateEnum::Focused, Color(255,0,0,128), Color(255,0,0,255) );
		m_pDebugOverlay = pDebugOverlay;
		m_afterglowFrames = 4;
	}

	RootPanel::RootPanel( GfxDevice * pGfxDevice ) : RootPanel()
	{
		if( pGfxDevice )
			m_geo = Size::fromPX( pGfxDevice->canvasSize() );
		m_pGfxDevice = pGfxDevice;
	}



	//____ Destructor _____________________________________________________________

	RootPanel::~RootPanel()
	{
	}

	//____ typeInfo() _________________________________________________________

	const TypeInfo& RootPanel::typeInfo(void) const
	{
		return TYPEINFO;
	}

	//____ setGfxDevice() _________________________________________________________

	bool RootPanel::setGfxDevice( GfxDevice * pDevice )
	{
		m_pGfxDevice = pDevice;

		if( m_pGfxDevice && !m_bHasGeo && slot._widget() )
			OO(slot._widget())->_resize( Size::fromPX( m_pGfxDevice->canvasSize() ) );

		m_dirtyPatches.add( geo() );
		return true;
	}

	//_____ setGeo() _____________________________________________________________

	bool RootPanel::setGeo( const Rect& geo )
	{
		m_geo = geo.aligned();

		if( geo.x == 0 && geo.y == 0 && geo.w == 0 && geo.h == 0 )
			m_bHasGeo = false;
		else
			m_bHasGeo = true;

		return true;
	}

	//____ geo() __________________________________________________________________

	Rect RootPanel::geo() const
	{
		if( m_bHasGeo )
			return m_geo;
		else if( m_pGfxDevice )
			return Rect( Coord(0,0), Size::fromPX( m_pGfxDevice->canvasSize() ) );
		else
			return Rect();
	}

	//____ _object() ____________________________________________________________

	Object * RootPanel::_object()
	{
		return this;
	}

	const Object * RootPanel::_object() const
	{
		return this;
	}

	//____ setVisible() ___________________________________________________________

	bool RootPanel::setVisible( bool bVisible )
	{
		if( bVisible != m_bVisible )
		{
			m_bVisible = bVisible;
			m_dirtyPatches.add( geo() );
		}
		return true;
	}

	//____ setDebugMode() ______________________________________________________

	void RootPanel::setDebugMode( bool onOff )
	{
		m_bDebugMode = onOff;

		// Make sure to clean up on the screen.

		if( m_bDebugMode == false )
		{
			for( auto it = m_afterglowRects.begin() ; it != m_afterglowRects.end() ; it++ )
				m_dirtyPatches.add( &(*it) );

			m_afterglowRects.clear();
		}
	}


	//____ setDebugOverlay() ____________________________________________________

	void RootPanel::setDebugOverlay( Skin * pOverlaySkin )
	{
		m_pDebugOverlay = pOverlaySkin;
	}

	//____ setDebugAfterglow() ____________________________________________________

	void RootPanel::setDebugAfterglow( int frames )
	{
		limit(frames, 0, 100);
		m_afterglowFrames = frames;
	}

	//____ setSkin() _____________________________________________________________

	void RootPanel::setSkin(Skin * pSkin)
	{
		m_pSkin = pSkin;
		m_dirtyPatches.clear();
		m_dirtyPatches.push(m_geo);
	}


	//____ render() _______________________________________________________________

	bool RootPanel::render()
	{
		return render( geo() );
	}

	bool RootPanel::render( const Rect& clip )
	{


		if( !beginRender() )
			return false;

		if( !renderSection(clip) )
		{
			endRender();
			return false;
		}

		if( !endRender() )
			return false;

		return true;
	}

	//____ beginRender() __________________________________________________________

	bool RootPanel::beginRender()
	{
		if( !m_pGfxDevice || !slot._widget() )
			return false;						// No GFX-device or no widgets to render.

		// Handle preRender calls.

		for( auto& pWidget : m_preRenderCalls )
			OO(pWidget)->_preRender();

		m_preRenderCalls.clear();

		// Handle debug overlays.

		if( m_bDebugMode )
		{
			// Remove from afterglow queue patches that are overlapped by our new dirty patches.

			for( std::deque<Patches>::iterator it = m_afterglowRects.begin() ; it != m_afterglowRects.end() ; ++it )
				it->sub(&m_dirtyPatches);

			// Add our new dirty patches to the top of the afterglow queue.

			m_afterglowRects.push_front(Patches());
			m_afterglowRects.front().add(&m_dirtyPatches);

			// Possibly remove overlays from the back, put them into dirty rects for re-render

			while( (int) m_afterglowRects.size() > m_afterglowFrames+1 )
			{
				m_dirtyPatches.add( &m_afterglowRects.back() );
				m_afterglowRects.pop_back();
			}

			// Re-render graphics behind overlays that go from state FOCUSED to NORMAL

			if( m_afterglowRects.size() > 1 )
			{
				m_dirtyPatches.add( &m_afterglowRects[1] );
			}
		}

		// Initialize GFX-device.

		return m_pGfxDevice->beginRender();
	}


	//____ renderSection() __________________________________________________________

	bool RootPanel::renderSection( const Rect& _clip )
	{
		if( !m_pGfxDevice || !slot._widget() )
			return false;						// No GFX-device or no widgets to render.

		// Make sure we have a vaild clip rectangle (doesn't go outside our geometry and has an area)

		Rect canvas = geo();
		Rect clip = Rect(_clip, canvas).aligned();
		if( clip.w == 0 || clip.h == 0 )
			return false;						// Invalid rect area.

		// Nothing to render if we are hidden

		if( !m_bVisible )
			return true;						// Not an error, just hidden.

		// Copy and clip our dirty patches

		Patches dirtyPatches( m_dirtyPatches, clip );

		// Render the dirty patches recursively

		if( dirtyPatches.size() > 0 )
		{
			ClipPopData clipPop = patchesToClipList(m_pGfxDevice, dirtyPatches);
			if (m_pSkin)
				m_pSkin->render(m_pGfxDevice, canvas, StateEnum::Normal);

			OO(slot._widget())->_render( m_pGfxDevice, canvas, canvas );
			popClipList(m_pGfxDevice, clipPop);
		}

		// Handle updated rect overlays

		if( m_bDebugMode && m_pDebugOverlay )
		{
			// Set clipping rectangle.

			RectI myClip = clip.px();

			m_pGfxDevice->setClipList(1, &myClip);

			// Render our new overlays

			for( const Rect * pRect = m_afterglowRects[0].begin() ; pRect != m_afterglowRects[0].end() ; pRect++ )
			{
				m_pDebugOverlay->render( m_pGfxDevice, *pRect, StateEnum::Focused );
			}

			// Render overlays that have turned into afterglow

			if( m_afterglowRects.size() > 1 )
			{
				for( const Rect * pRect = m_afterglowRects[1].begin() ; pRect != m_afterglowRects[1].end() ; pRect++ )
				{
					m_pDebugOverlay->render( m_pGfxDevice, *pRect, StateEnum::Normal );
				}
			}
		}

		return true;
	}

	//____ endRender() ____________________________________________________________

	bool RootPanel::endRender( void )
	{
		if( !m_pGfxDevice || !slot._widget() )
			return false;						// No GFX-device or no widgets to render.

		// Turn dirty patches into update patches
		//TODO: Optimize by just making a swap.

		m_updatedPatches.clear();
		m_updatedPatches.add(&m_dirtyPatches);
		m_dirtyPatches.clear();

		return m_pGfxDevice->endRender();
	}


	//____ _findWidget() _____________________________________________________________

	Widget * RootPanel::_findWidget( const Coord& ofs, SearchMode mode )
	{
		if( !geo().contains(ofs) || !slot._widget() )
			return 0;

		if(slot._widget() && slot._widget()->isContainer() )
			return OO(static_cast<Container*>(slot._widget()))->_findWidget( ofs, mode );

		return slot._widget();
	}


	//____ _focusedChild() ______________________________________________________

	Widget * RootPanel::_focusedChild() const
	{
		if( !m_pFocusedChild )
			return slot._widget();

		return m_pFocusedChild;
	}


	//____ _childPos() ________________________________________________________

	Coord RootPanel::_childPos( const StaticSlot * pSlot ) const
	{
		return geo().pos();
	}

	//____ _childGlobalPos() __________________________________________________

	Coord RootPanel::_childGlobalPos( const StaticSlot * pSlot ) const
	{
		return geo().pos();
	}

	//____ _isChildVisible() __________________________________________________

	bool RootPanel::_isChildVisible( const StaticSlot * pSlot ) const
	{
		return true;
	}

	//____ _childWindowSection() __________________________________________________

	Rect RootPanel::_childWindowSection( const StaticSlot * pSlot ) const
	{
		return geo();
	}

	//____ _slotTypeInfo() ________________________________________________________

	const TypeInfo&	RootPanel::_slotTypeInfo(const StaticSlot * pSlot) const
	{
		return DynamicSlot::TYPEINFO;
	}

	//____ _container() __________________________________________________

	Container * RootPanel::_container()
	{
		return nullptr;
	}

	//____ _root() __________________________________________________

	RootPanel * RootPanel::_root()
	{
		return this;
	}

	//____ _childRequestRender() __________________________________________________

	void RootPanel::_childRequestRender( StaticSlot * pSlot )
	{
		if( m_bVisible )
			m_dirtyPatches.add( geo() );
	}

	void RootPanel::_childRequestRender( StaticSlot * pSlot, const Rect& rect )
	{
		if( m_bVisible )
			m_dirtyPatches.add( Rect( geo().pos() + rect.pos(), rect.size() ) );
	}

	//____ _childRequestResize() __________________________________________________

	void RootPanel::_childRequestResize( StaticSlot * pSlot )
	{
		OO(slot)._setSize(m_geo.size());
	}

	//____ _childRequestFocus() __________________________________________________

	bool RootPanel::_childRequestFocus( StaticSlot * pSlot, Widget * pWidget )
	{
		if( pWidget == m_pFocusedChild.rawPtr() )
			return true;

		Widget * pOldFocus = m_pFocusedChild.rawPtr();
		m_pFocusedChild = pWidget;
		return Base::inputHandler()->_focusChanged( this, pOldFocus, pWidget );
	}

	//____ _childReleaseFocus() __________________________________________________

	bool RootPanel::_childReleaseFocus( StaticSlot * pSlot, Widget * pWidget )
	{
		if( pWidget != m_pFocusedChild.rawPtr() )
			return true;					// Never had focus, although widget seems to believe it.

		if( pWidget == slot._widget() )
			return false;

		Widget * pOldFocus = m_pFocusedChild;
		m_pFocusedChild = slot._widget();
		return Base::inputHandler()->_focusChanged( this, pOldFocus, slot._widget());
	}

	//____ _childRequestInView() __________________________________________________

	void RootPanel::_childRequestInView( StaticSlot * pSlot )
	{
		// Do nothing, root ignores inView requests.
	}
	void RootPanel::_childRequestInView( StaticSlot * pSlot, const Rect& mustHaveArea, const Rect& niceToHaveArea )
	{
		// Do nothing, root ignores inView requests.
	}

	//____ _prevChild() __________________________________________________

	Widget * RootPanel::_prevChild( const StaticSlot * pSlot ) const
	{
		return nullptr;
	}

	//____ _nextChild() __________________________________________________

	Widget * RootPanel::_nextChild( const StaticSlot * pSlot ) const
	{
		return nullptr;
	}

	//____ _releaseChild() ____________________________________________________

	void RootPanel::_releaseChild(StaticSlot * pSlot)
	{
		_replaceChild(pSlot, nullptr);
	}

	//____ _replaceChild() ____________________________________________________________

	void RootPanel::_replaceChild(StaticSlot * pSlot, Widget * pNewWidget)
	{
		Rect myGeo = geo();

		if( slot._widget())
			OO(slot._widget())->_collectPatches(m_dirtyPatches, myGeo, myGeo);

		OO(slot)._setWidget(pNewWidget);

		if (pNewWidget)
		{
			OO(pNewWidget)->_resize(myGeo.size());
			OO(pNewWidget)->_collectPatches(m_dirtyPatches, myGeo, myGeo);
		}
	}

	//____ _selectSlots() _____________________________________________________

	void RootPanel::_selectSlots(StaticSlot * pSlot, int nb)
	{
		return;				// RootPanel doesn't support selection
	}

	//____ _unselectSlots() ___________________________________________________

	void RootPanel::_unselectSlots(StaticSlot * pSlot, int nb)
	{
		return;				// RootPanel doesn't support selection
	}

	//____ _repadSlots() ______________________________________________________

	void RootPanel::_repadSlots(StaticSlot * pSlot, int nb, Border padding)
	{
		return;				// RootPanel doesn't support padding
	}

	void RootPanel::_repadSlots(StaticSlot * pSlot, int nb, const Border * pPadding)
	{
		return;				// RootPanel doesn't support padding
	}

	//____ _didAddSlots() _____________________________________________________

	void RootPanel::_didAddSlots(StaticSlot * pSlot, int nb)
	{
		return;				// RootPanel doesn't support adding/moving/erasing slots
	}

	//____ _didMoveSlots() ____________________________________________________

	void RootPanel::_didMoveSlots(StaticSlot * pFrom, StaticSlot * pTo, int nb)
	{
		return;				// RootPanel doesn't support adding/moving/erasing slots
	}

	//____ _willEraseSlots() __________________________________________________

	void RootPanel::_willEraseSlots(StaticSlot * pSlot, int nb)
	{
		return;				// RootPanel doesn't support adding/moving/erasing slots
	}

	//____ _hideSlots() _______________________________________________________

	void RootPanel::_hideSlots(StaticSlot * pSlot, int nb)
	{
		return;				// RootPanel doesn't support hiding
	}

	//____ _unhideSlots() _____________________________________________________

	void RootPanel::_unhideSlots(StaticSlot * pSlot, int nb)
	{
		return;				// RootPanel doesn't support hiding
	}

} // namespace wg
